"""Class to describe a portion of a spreadsheet"""
from __future__ import annotations

from dataclasses import dataclass

from . import operations
from .custom_types import Cells, CellsRange, CoordinatesPair


@dataclass
class SpreadsheetElement:
    """
    Class to describe a rectangular portion of a spreadsheet.

    Identify a rectangular part of a spreadsheet by providing a pair of
    coordinates that represent the top left and bottom right corner of the
    rectangle.

    This object allows to access the various forms in which the rectangle
    of cells can be identified (cells, coordinates, cells_range).

    Parameters
    ----------
    coordinates : tuple[tuple[int, int], tuple[int, int]]
        A tuple containing two tuples of length two, each containing two numbers
        which univocally identify a cell. The two numbers are indexes of the
        respectively the columns and rows. The two cells are the top left and
        bottom right one.
    """

    coordinates: CoordinatesPair

    @property
    def cells(self) -> Cells:
        """
        Return tuple of cells making up the object in the form ("A1", "A2").
        """
        return operations.cells_rectangle(self.coordinates)

    @property
    def cells_range(self) -> CellsRange:
        """
        Return a str giving info on the top left and bottom right cells of the
        object, in the form "A1:B3".
        """
        return operations.cells_range(self.coordinates)



"""Functions translating positional information of spreadsheet's cells across
several forms.

The most basic notation is the coordinates system. Analogously to a cartesian
plane, it identifies each cell of the spreadsheet through its position defined
by a (x, y) pair. The origin is the cell 'A1', the top left corner, that
corresponds to coordinates (0, 0). Moving left and downwards increase
respectively the first and second coordinate. So that 'B1' is (1, 0) and
'A2' is (0, 1).
"""
from __future__ import annotations

import string
from itertools import product

from .custom_types import Cell, Cells, CellsRange, Coordinates, CoordinatesPair


def cells_rectangle(coordinates_pair: CoordinatesPair) -> Cells:
    """Return a tuple listing the labels of the cells contained in a rectangular
    portion of a spreadsheet.

    The rectangle is identified by the coordinates of its top left and bottom
    right corner cells.

    Parameters
    ----------
    coordinates_pair : tuple[tuple[int, int], tuple[int, int]]
        Tuple cotaining two coordinates, at their time represented as tuple
        containing two integers. They identify respectively the top left and
        bottom right cell of a rectangular portion of a spreadsheet.

    Returns
    -------
    tuple[str, ...]
        Tuple in the form ('A1', 'A2', ...).

    Examples
    -------
    >>> from spreadpandas.operations import cells_rectangle
    >>> coords_pair = ((0, 0), (1, 1))
    >>> cells_rectangle(coords_pair)
    ('A1', 'B1', 'A2', 'B2')
    """

    starting_letter_pos, starting_number = coordinates_pair[0]
    ending_letter_pos, ending_number = coordinates_pair[1]

    prod = product(
        range(starting_number, ending_number + 1),
        range(starting_letter_pos, ending_letter_pos + 1),
    )

    return tuple(cell_from_coordinates((letter, number)) for number, letter in prod)


def cells_range(coordinates_pair: CoordinatesPair) -> CellsRange:
    """Return a string in the form "A1:B2" given a pair of coordinates in the
    form ((0,0),(1,1)).

    Parameters
    ----------
    coordinates_pair : tuple[tuple[int, int], tuple[int, int]]
        Tuple cotaining two coordinates, at their time represented as tuple
        containing two integers.

    Returns
    -------
    str
        String in the form "A1:B2".
    """
    cell1 = cell_from_coordinates(coordinates_pair[0])
    cell2 = cell_from_coordinates(coordinates_pair[1])
    return f"{cell1}:{cell2}"


def cell_from_coordinates(coordinates: Coordinates) -> Cell:
    """Return the string representation of a cell in the form "A1" given its
    coordinates in the form (0, 0).

    Parameters
    ----------
    coordinates : tuple[int, int]
        A tuple of two integers, the first representing the column index
        and the second the row index.

    Returns
    -------
    str
        Label of the cell in the form "A1".
    """
    return letter_from_index(coordinates[0]) + row_number_from_index(coordinates[1])


def row_number_from_index(row_index: int) -> str:
    """Return the number that labels the row given its index. The convertion
    simply consists in adding 1 and converting to string.

    Spreadsheets label rows starting from 1, so that their number is 1 more
    than its zero-indexed position.

    Parameters
    ----------
    row_index : int
        Index of the row to convert.

    Returns
    -------
    str
        Label as conceived by spreadsheets to identify the row.
    """
    if row_index < 0:
        raise ValueError(f"Negative values are not accepted: {row_index} was passed.")
    return str(row_index + 1)


def letter_from_index(letter_index: int) -> str:
    """
    Returns the spreadsheet column's letter given index; ex: 0 -> "A", 26 -> "AA"

    ------------------
    The index should be interpreted as the place where the column is
    counting from left to right. The count starts from 0, which corresponds
    to "A", 1 to "B" and so on. The current program handles the indexes up to
    18 277, corresponding to column "ZZZ".
    """
    if letter_index < 0:
        raise ValueError(f"Minimum accepted value is 0, {letter_index} was provided")

    units_letter = string.ascii_uppercase[letter_index % 26]
    if letter_index <= 25:
        return units_letter

    hundreds_letter = string.ascii_uppercase[((letter_index - 26) % (26**2)) // 26]
    if letter_index <= (26**2 + 25):
        return hundreds_letter + units_letter

    thousands_letter = string.ascii_uppercase[
        (letter_index - 26**2 - 26) % (26**3) // 26**2
    ]
    if letter_index <= (26**3 + 26**2 + 25):
        return thousands_letter + hundreds_letter + units_letter

    raise ValueError("The program does not handle indexes past 18 277 yet")


def index_from_letter(spreadsheet_letter: str) -> int:
    """
    Returns the col index given spreadsheet letter; ex: "A" -> 0, "AA" -> 26

    ------------------
    The spreadsheet letter is the column label used in spreadsheets to identify
    the column. The first one is "A" which corresponds to 0. The current program
    handles the columns up to "ZZZ", corresponding to number 18 277.
    """
    units_letter = string.ascii_uppercase.index(spreadsheet_letter[-1])
    if len(spreadsheet_letter) == 1:
        return units_letter

    hundreds_letter = string.ascii_uppercase.index(spreadsheet_letter[-2]) + 1
    if len(spreadsheet_letter) == 2:
        return 26 * hundreds_letter + units_letter

    thousands_letter = string.ascii_uppercase.index(spreadsheet_letter[-3]) + 1
    if len(spreadsheet_letter) == 3:
        return 26**2 * thousands_letter + 26 * hundreds_letter + units_letter

    raise ValueError("The program does not handle column letters past 'ZZZ' yet")



from setuptools import find_packages, setup

with open("readme.md", "r") as file:
    long_description = file.read()

setup(
    name="spreadpandas",
    version="0.0.1",
    description="Get the spreadsheet representation of a pandas data frame.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FilippoPisello/SpreadPandas",
    author="Filippo Pisello",
    author_email="filippo.pisello@live.it",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=["numpy>=1.10", "pandas>=1.1.0"],
    python_requires=">=3.7",
)



"""Main class to represent a Spreadsheet object"""
from __future__ import annotations

import pandas as pd

from .spreadsheet_element import SpreadsheetElement


class Spreadsheet:
    """
    Class to represent a pandas dataframe to be loaded into a spreadsheet.

    The class intakes as main argument a pandas dataframe. Based on the dimensions,
    it finds a correspondence between the dataframe and its representation on a
    spreadsheet. The attributes are then the cells that the various parts of
    the dataframe would occupy on a spreadsheet.

    The spreadsheet elements are the following.
    - Header: cells which contain the column names. Can be of depth greater than
      one in case of multicolumns.
    - Index: cells which contain the row names. Can be of depth greater than
      one in case of multiindex.
    - Body: cells which contain the actual data.
    - Table: the union of all the cells above.

    If no rows or columns are skipped, it is assumed that the content is loaded
    into the spreadsheet starting from the top left cell, A1.
    """

    def __init__(
        self,
        dataframe: pd.DataFrame,
        keep_index: bool = False,
        keep_header: bool = True,
        skip_rows: int = 0,
        skip_columns: int = 0,
    ):
        """Create a Spreadsheet object to get the dimensionality of a pandas
        data frame when exported to a spreadsheet.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Pandas data frame to be represented as spreadsheet.
        keep_index : bool, optional
            If True, index is considered as an element to be mapped into the
            spreadsheet. If False, it is ignored, by default False.
        keep_header : bool, optional
            If True, header is considered as an element to be mapped into the
            spreadsheet. If False, it is ignored, by default True.
        skip_rows : int, optional
            The number of rows to skip - left empty - at the top of the
            spreadsheet. Data will be placed starting from row skip_rows + 1.
            By default 0.
        skip_columns : int, optional
            The number of columns to skip - left empty- at the left of the
            spreadsheet. Data will be placed starting from column
            skip_columns + 1. By default 0.
        """
        self.keep_index = keep_index
        self.keep_header = keep_header
        self.df = dataframe
        self.skip_rows = skip_rows
        self.skip_cols = skip_columns

    # --------------------------------
    # Data frame derived properties
    # --------------------------------
    @property
    def depth_index(self):
        return self.df.index.nlevels * self.keep_index

    @property
    def depth_columns(self):
        return self.df.columns.nlevels * self.keep_header

    # --------------------------------
    # SpreadsheetElements
    # --------------------------------
    @property
    def body(self) -> SpreadsheetElement:
        start_col = self.depth_index + self.skip_cols
        start_row = self.depth_columns + self.skip_rows
        end_col = start_col - 1 + self.df.shape[1]
        end_row = start_row - 1 + self.df.shape[0]
        return SpreadsheetElement(((start_col, start_row), (end_col, end_row)))

    @property
    def header(self) -> SpreadsheetElement:
        if not self.keep_header:
            return None

        start_row = self.skip_rows
        end_row = start_row + self.depth_columns - 1
        return SpreadsheetElement(
            (
                (self.body.coordinates[0][0], start_row),
                (self.body.coordinates[1][0], end_row),
            )
        )

    @property
    def index(self) -> SpreadsheetElement | None:
        if not self.keep_index:
            return None

        start_col = self.skip_cols
        end_col = start_col + self.depth_index - 1
        return SpreadsheetElement(
            (
                (start_col, self.body.coordinates[0][1]),
                (end_col, self.body.coordinates[1][1]),
            )
        )

    @property
    def table(self) -> SpreadsheetElement:
        if self.keep_index:
            start_col = self.index.coordinates[0][0]
        else:
            start_col = self.body.coordinates[0][0]

        if self.keep_header:
            start_row = self.header.coordinates[0][1]
        else:
            start_row = self.body.coordinates[0][1]

        return SpreadsheetElement(
            (
                (start_col, start_row),
                (self.body.coordinates[1][0], self.body.coordinates[1][1]),
            )
        )



"""Module where custom types are defined"""
from __future__ import annotations

from typing import Tuple

Coordinates = Tuple[int, int]
CoordinatesPair = Tuple[Coordinates, Coordinates]
Cell = str
Cells = Tuple[Cell, ...]
CellsRange = str
